<?php
/**
 * Gestion du formulaire de chargement ou de vidage des tables de codes ISO.
 *
 * @package    SPIP\ISOCODE\OBJET
 */
if (!defined('_ECRIRE_INC_VERSION')) {
	return;
}

/**
 * Chargement des données : le formulaire propose les actions possibles sur les tables de codes ISO,
 * à savoir, charger ou vider et la liste des tables regroupées par service.
 * L'action vider s'appelle décharger car il existe dékà une fonction d'administration de vidage des tables.
 *
 * @param string $type Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 *
 * @return array Tableau des données à charger par le formulaire (affichage). Aucune donnée chargée n'est un
 *               champ de saisie, celle-ci sont systématiquement remises à zéro.
 *               - `_actions_tables`  : (affichage) alias et libellés des actions possibles sur une table, `charger` et `decharger`
 *               - `_actions_disable` : (affichage) liste des actions désactivées (`decharger` si aucune table n`est chargée)
 *               - `_action_defaut`   : (affichage) action sélectionnée par défaut, `charger`
 *               - `_tables`          : (affichage) noms des tables sans le préfixe `spip_`
 */
function formulaires_isocode_peupler_charger(string $type) : array {
	$valeurs = [];

	// Lister les actions sur les tables
	include_spip('inc/isocode');
	$valeurs['_action_explication'] = _T("isocode:explication_action_{$type}");
	$valeurs['_actions'] = [
		'charger'   => _T('isocode:label_action_charger'),
		'decharger' => _T('isocode:label_action_decharger')
	];

	// Construire la liste des éléments et leur statut de chargement
	$valeurs['_label_element'] = _T("isocode:label_element_{$type}");
	$aucune_element_charge = true;
	if ($type === 'nomenclature') {
		$tables = isocode_lister_tables($type, true);
		if ($tables) {
			foreach ($tables as $_groupe => $_tables) {
				$groupe = _T("isocode:label_groupe_{$_groupe}");
				foreach ($_tables as $_table) {
					$valeurs['_elements'][$groupe][$_table] = "<em>{$_table}</em>, " . _T("isocode:{$type}_{$_table}");
					if (isocode_lire_consignation($type, '', $_table)) {
						$valeurs['_elements'][$groupe][$_table] .= ' - [' . _T('isocode:info_charge') . ']';
						$aucune_element_charge = false;
					}
				}
			}
		}
	} else {
		$services = isocode_lister_services($type);
		foreach ($services as $_service) {
			$valeurs['_elements'][$_service] = _T("isocode:{$type}_{$_service}");
			if (isocode_lire_consignation($type, $_service, '')) {
				$valeurs['_elements'][$_service] .= ' - [' . _T('isocode:info_charge') . ']';
				$aucune_element_charge = false;
			}
		}
	}

	// Désactiver l'action vider si aucun table n'est chargée
	if ($aucune_element_charge) {
		$valeurs['_actions_disable'] = ['decharger' => 'oui'];
		$valeurs['_action_defaut'] = 'charger';
	}

	return $valeurs;
}

/**
 * Vérification des saisies : il est indispensable de choisir une action (`decharger` ou `charger`) et
 * une table.
 *
 * @param string $type Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 *
 * @return array Tableau des erreurs sur l'action et/ou la table ou tableau vide si aucune erreur.
 */
function formulaires_isocode_peupler_verifier(string $type) : array {
	$erreurs = [];

	$obligatoires = ['action_peuplement', 'elements'];
	foreach ($obligatoires as $_obligatoire) {
		if (!_request($_obligatoire)) {
			$erreurs[$_obligatoire] = _T('info_obligatoire');
		}
	}

	return $erreurs;
}

/**
 * Exécution du formulaire : les tables choisies sont soit vidées, soit chargées.
 *
 * @uses isocode_charger_tables()
 * @uses isocode_decharger_tables()
 * @uses formater_message()
 *
 * @param string $type Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 *
 * @return array Tableau retourné par le formulaire contenant toujours un message de bonne exécution ou
 *               d'erreur. L'indicateur editable est toujours à vrai.
 */
function formulaires_isocode_peupler_traiter(string $type) : array {
	// Acquisition des saisies: comme elles sont obligatoires, il existe toujours une action et une
	// table ou un service suivant le type de service.
	$action = _request('action_peuplement');
	$elements = _request('elements');

	// On construit le tableau services/tables à fournir aux actions charger ou décharger.
	// -- tableau de service => [tables]
	include_spip('inc/isocode');
	$services = [];
	if ($type === 'nomenclature') {
		foreach ($elements as $_table) {
			$service = isocode_trouver_service($type, $_table);
			$services[$service[0]][] = $_table;
		}
	} else {
		foreach ($elements as $_service) {
			$table = isocode_trouver_table($type, $_service);
			$services[$_service][] = $table[0];
		}
	}

	// Pour chaque couple (service, table) on génère l'action demandée.
	// (La fonction de chargement de la table lance un vidage préalable si la table demandé est déjà chargée)
	$statuts = [];
	foreach ($services as $_service => $_tables) {
		foreach ($_tables as $_table) {
			$actionner = "isocode_{$action}";
			$statuts[] = $actionner($type, $_service, $_table);
		}
	}

	// Formatage du message avant renvoie au formulaire
	$retour = isocode_peupler_notifier($type, $action, $statuts);
	$retour['editable'] = true;

	return $retour;
}

/**
 * Formate les messages de succès et d'erreur résultant des actions de chargement ou de vidage
 * des tables de codes ISO.
 *
 * @param string $type    Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string $action  Action venant d'être appliquée à certaines tables. Peut prendre les valeurs `charger` et
 *                        `decharger`.
 * @param array  $statuts Tableau résultant de l'action sur les tables choisies:
 *                        - `erreur`  : identifiant de l'erreur, à savoir, `ok`, `nok` ou `sha`.
 *                        - `type`    : type de service.
 *                        - `service` : identifiant du service.
 *                        - `table`   : nom de la table sans le préfixe spip_.
 *
 * @return array Tableau des messages à afficher sur le formulaire:
 *               - `message_ok`     : message sur les tables ayant été traitées avec succès ou tableau vide sinon.
 *               - `message_erreur` : message sur les tables en erreur ou tableau vide sinon.
 */
function isocode_peupler_notifier(string $type, string $action, array $statuts) : array {
	$messages = [
		'message_ok'     => '',
		'message_erreur' => ''
	];
	$variables = [
		'ok'  => [],
		'nok' => [],
		'sha' => [],
	];
	$statut_global = [
		'ok'  => false,
		'nok' => false,
		'sha' => false,
	];

	// On compile la liste des pays traités et un indicateur global pour chaque cas d'erreur.
	foreach ($statuts as $_statut) {
		// Traitement des succès
		if ($_statut['erreur'] === 'sha') {
			$variables['sha'][] = ($type === 'nomenclature') ? $_statut['table'] : $_statut['service'];
			$statut_global['sha'] = true;
		} elseif ($_statut['erreur'] === 'nok') {
			$variables['nok'][] = ($type === 'nomenclature') ? $_statut['table'] : $_statut['service'];
			$statut_global['nok'] = true;
		} else {
			$variables['ok'][] = ($type === 'nomenclature') ? $_statut['table'] : $_statut['service'];
			$statut_global['ok'] = true;
		}
	}

	// Traitement des succès
	if ($statut_global['ok']) {
		$messages['message_ok'] .= _T("isocode:succes_{$action}", ['elements' => implode(', ', $variables['ok'])]);
	}

	// Traitement des erreurs
	if ($statut_global['nok']) {
		$messages['message_erreur'] .= _T("isocode:erreur_{$action}", ['elements' => implode(', ', $variables['nok'])]);
	}
	if ($statut_global['sha']) {
		$messages['message_erreur'] .= $messages['message_erreur'] ? '<br />' : '';
		$messages['message_erreur'] .= _T("isocode:notice_{$action}", ['elements' => implode(', ', $variables['sha'])]);
	}

	return $messages;
}
