<?php
/**
 * Ce fichier contient l'ensemble des fonctions implémentant l'API du plugin.
 *
 * @package SPIP\ISOCODE\API
 */
if (!defined('_ECRIRE_INC_VERSION')) {
	return;
}

/**
 * Charge en base de données une table de nomenclatures peuplée par un service donné.
 *
 * @api
 *
 * @uses lire_source()
 * @uses isocode_decharger()
 * @uses inserer_enregistrements()
 * @uses consigner_chargement()
 *
 * @param string $type    Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string $service Identifiant du service.
 * @param string $table   Nom de la table sans le préfixe spip_.
 *
 * @throws Exception
 *
 * @return array Tableau associatif résultat de l'action de chargement:
 *               - `erreur`  : identifiant de l'erreur, à savoir, `ok`, `nok` ou `sha`.
 *               - `type`    : type de service.
 *               - `service` : identifiant du service.
 *               - `table`   : nom de la table sans le préfixe spip_.
 */
function isocode_charger(string $type, string $service, string $table) : array {
	$retour = [
		'erreur'  => '',
		'type'    => $type,
		'service' => $service,
		'table'   => $table,
	];

	// Lecture des données contenues soit dans un fichier, soit dans une page web, soit dans le retour d'une API REST
	// et renvoie une liste d'éléments prêts à être enregistrés dans la table concernée.
	// Si la table est déjà chargée et que le fichier ou la page source n'a pas changé, la fonction de
	// lecture ne renvoie aucun élément pour éviter des traitements inutiles mais renvoie un indicateur
	// sur le SHA.
	$erreur = true;

	spip_timer('lire');
	include_spip('inc/isocode_utils');
	[$enregistrements, $sha, $source_identique, $groupe] = lire_source($type, $service, $table);
	$duree_lire = spip_timer('lire');
	if ($enregistrements) {
		spip_timer('ecrire');
		// Suppression des éléments éventuels déjà chargés. On ne gère pas d'erreur
		// sur ce traitement car elle sera forcément détectée lors de l'insertion qui suit.
		isocode_decharger($type, $service, $table);

		// Insertion dans la base de données des éléments extraits
		if (inserer_enregistrements($type, $table, $enregistrements)) {
			// On stocke les informations de chargement de la table dans une meta.
			$meta = [
				'sha'    => $sha,
				'nbr'    => count($enregistrements),
				'maj'    => date('Y-m-d H:i:s'),
				'groupe' => $groupe
			];
			consigner_chargement($meta, $type, $service, $table);
			$erreur = false;
		}
		$duree_ecrire = spip_timer('ecrire');
		spip_log("La table <{$table}> a été chargée via le service <{$service}> (lecture source: {$duree_lire} - insertion BD: {$duree_ecrire})", 'isocode' . _LOG_DEBUG);
	}

	// Si la table est en erreur, on passe l'indicateur global à erreur et on stocke la table en nok
	// ou en sha identique suivant la cas.
	// Si le traitement est ok on stocke juste la table.
	if ($erreur) {
		if ($source_identique) {
			$retour['erreur'] = 'sha';
		} else {
			$retour['erreur'] = 'nok';
		}
	}

	return $retour;
}

/**
 * Supprime en base de données, le contenu d'une table de nomenclatures peuplée par un service donné.
 * La meta concernant les informations de chargement de chaque table est aussi effacée.
 *
 * @api
 *
 * @uses deconsigner_chargement()
 *
 * @param string $type    Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string $service Identifiant du service.
 * @param string $table   Nom de la table sans le préfixe spip_.
 *
 * @return array Tableau associatif résultat de l'action de déchargement:
 *               - `erreur`  : identifiant de l'erreur, à savoir, `ok`, `nok` ou `sha`.
 *               - `type`    : type de service.
 *               - `service` : identifiant du service.
 *               - `table`   : nom de la table sans le préfixe spip_.
 */
function isocode_decharger(string $type, string $service, string $table) : array {
	$retour = [
		'erreur'  => '',
		'type'    => $type,
		'service' => $service,
		'table'   => $table,
	];

	$where = ($type === 'nomenclature')
		? []
		: ['service=' . sql_quote($service)];
	$sql_ok = sql_delete("spip_{$table}", $where);
	if ($sql_ok !== false) {
		// Supprimer la meta propre à la table.
		include_spip('inc/isocode_utils');
		deconsigner_chargement($type, $service, $table);
	} else {
		$retour['erreur'] = 'nok';
	}

	return $retour;
}

/**
 * Retourne la liste des services disponibles pour le chargement des tables de nomenclatures.
 * La fonction lit les sous-répertoires du répertoire `services/` du plugin dont le nom correspond au type de service.
 *
 * @api
 *
 * @return array La liste des services disponibles ou tableau vide si aucun service n'est détecté.
 */
function isocode_lister_types_service() : array {
	static $types = [];

	if (!$types) {
		if ($dossiers = glob(_DIR_PLUGIN_ISOCODE . 'services/*', GLOB_ONLYDIR)) {
			foreach ($dossiers as $_dossier) {
				$types[] = strtolower(basename($_dossier));
			}
		}
	}

	return $types;
}

/**
 * Retourne la liste des services disponibles pour un type de service donné.
 *
 * @api
 *
 * @param string $type Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 *
 * @return array La liste des services disponibles ou tableau vide aucun service n'est détecté.
 */
function isocode_lister_services(string $type) : array {
	static $services = [];

	if (
		$type
		and !isset($services[$type])
	) {
		if ($type === 'nomenclature') {
			// Pour ce type de service, chaque service est dans un sous-dossier au nom du service.
			// Ce sous-dossier ne contient donc qu'un service, il suffit de vérifier l'existence d'un fichier d'api.
			if ($dossiers = glob(_DIR_PLUGIN_ISOCODE . "services/{$type}/*", GLOB_ONLYDIR)) {
				foreach ($dossiers as $_dossier) {
					$service = strtolower(basename($_dossier));
					if (file_exists(_DIR_PLUGIN_ISOCODE . "services/{$type}/{$service}/{$service}_api.php")) {
						$services[$type][] = $service;
					}
				}
			}
		} else {
			// Contours géométriques et infrasubdivisions :
			// Les services sont les clés de la configuration incluse dans le fichier unique d'api.
			include_spip("services/{$type}/{$type}_api");
			$services[$type] = array_keys($GLOBALS['isocode'][$type]);
		}
	}

	return $services[$type] ?? [];
}

/**
 * Retourne la liste de toutes les tables gérées par un type de service donné.
 * Il est possible, pour le type de service `nomenclature` de demander le regourpement des tables selon une
 * logique de groupe définie dans la configuration (affichage uniquement).
 *
 * @api
 *
 * @uses isocode_lister_services()
 *
 * @param string    $type         Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param null|bool $avec_groupes Indique si la liste des tables doit être regroupée par groupe.
 *
 * @return array Liste des tables sans le préfixe `spip_`.
 */
function isocode_lister_tables(string $type, ?bool $avec_groupes = false) : array {
	$tables = [];

	// On vérifie l'argument type de service qui est vide si tous les types sont requis.
	if ($type) {
		$services = isocode_lister_services($type);
		if ($type === 'nomenclature') {
			// On collecte pour chaque service, la liste des tables qu'il supporte.
			foreach ($services as $_service) {
				include_spip("services/{$type}/{$_service}/{$_service}_api");
				if (!$avec_groupes) {
					$tables = array_merge($tables, array_keys($GLOBALS['isocode'][$_service]));
				} else {
					foreach ($GLOBALS['isocode'][$_service] as $_table => $_config) {
						$groupe = $_config['groupe'];
						$tables[$groupe][] = $_table;
					}
				}
			}
		} else {
			// Un service ne concerne qu'une seule table
			include_spip("services/{$type}/{$type}_api");
			foreach ($services as $_service) {
				$tables[] = $GLOBALS['isocode'][$type][$_service]['table'];
			}
			$tables = array_unique($tables);
		}
	}

	return $tables;
}

/**
 * Détermine le ou les services associés au chargement de la table choisie.
 *
 * @api
 *
 * @uses isocode_lister_services()
 * @uses isocode_lister_tables()
 *
 * @param string $type  Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string $table Nom d'une table sans le préfixe `spip_`.
 *
 * @return array Le ou les services associés à la table fournie suivant le type de service.
 *               - nomenclature     : tableau du service unique permettant le chargement de la table ou tableau vide.
 *               - geometrie        : tableau des services permettant le chargement de la table ou tableau vide.
 *               - infrasubdivision : tableau des services permettant le chargement de la table ou tableau vide.
 */
function isocode_trouver_service(string $type, string $table) : array {
	static $tables = [];
	static $services = [];
	$services_trouves = [];

	if ($type) {
		if (!isset($services[$type])) {
			$services[$type] = isocode_lister_services($type);
		}
		if (!isset($tables[$type])) {
			$tables[$type] = isocode_lister_tables($type);
		}

		if (in_array($table, $tables[$type])) {
			if ($type === 'nomenclature') {
				foreach ($services[$type] as $_service) {
					include_spip("services/{$type}/{$_service}/{$_service}_api");
					if (isset($GLOBALS['isocode'][$_service][$table])) {
						$services_trouves[] = $_service;
						break;
					}
				}
			} else {
				// Détermination du ou des services pour le type géométrie
				include_spip("services/{$type}/{$type}_api");
				foreach ($services[$type] as $_service) {
					if ($GLOBALS['isocode'][$type][$_service]['table'] === $table) {
						$services_trouves[] = $_service;
					}
				}
			}
		}
	}

	return $services_trouves;
}

/**
 * Détermine la ou les tables pouvant être peuplées par un service.
 *
 * @api
 *
 * @uses isocode_lister_services()
 *
 * @param string $type    Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string $service Identifiant du service.
 *
 * @return array Le ou les tables associées au service fourni suivant le type de service.
 *               - nomenclature     : tableau des services permettant le chargement de la table ou tableau vide.
 *               - geometrie        : tableau du service unique permettant le chargement de la table ou tableau vide.
 *               - infrasubdivision : tableau du service unique permettant le chargement de la table ou tableau vide.
 */
function isocode_trouver_table(string $type, string $service) : array {
	static $services = [];
	$tables_trouvees = [];

	if ($type) {
		if (!isset($services[$type])) {
			$services[$type] = isocode_lister_services($type);
		}

		if (in_array($service, $services[$type])) {
			if ($type === 'nomenclature') {
				include_spip("services/{$type}/{$service}/{$service}_api");
				$tables_trouvees = array_keys($GLOBALS['isocode'][$service]);
			} else {
				include_spip("services/{$type}/{$type}_api");
				$tables_trouvees[] = $GLOBALS['isocode'][$type][$service]['table'];
			}
		}
	}

	return $tables_trouvees;
}

/**
 * Renvoie la configuration des crédits d'un service.
 *
 * @api
 *
 * @param string      $type    Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string      $service Identifiant du service.
 * @param null|string $table   Nom de la table sans le préfixe spip_.
 *
 * @return array Tableau des données de crédits.
 */
function isocode_crediter(string $type, string $service, ?string $table = '') : array {
	include_spip('inc/isocode_utils');
	$credits = configuration_lire($type, $service, $table, 'credits');

	return $credits;
}

/**
 * Indique si une table est déjà chargée ou pas en base de données.
 * La fonction scrute la table `spip_${table}` et non la meta propre à la table.
 *
 * @api
 *
 * @param string $type    Type de service. Prend les valeurs `nomenclature`, `infra_subdivision` ou `geometrie`.
 * @param string $service Identifiant du service.
 * @param string $table   Nom de la table sans le préfixe spip_.
 *
 * @return array Tableau de la meta de consignation.
 */
function isocode_lire_consignation(string $type, string $service, string $table) : array {
	// Récupérer la meta propre au règne afin de la retourner.
	include_spip('inc/config');
	$meta = ($type === 'nomenclature')
		? lire_config("isocode/{$type}/{$table}", [])
		: lire_config("isocode/{$type}/{$service}", []);

	return $meta;
}
